
module AVG_CS_CONTROLLER 

	//////////////////PARAMETERS///////////////////////

#(
	parameter 
	DATA_SIZE_BITS 		 = 16, 
	PRECISION_BITS_MEMORY = 8, 
	CHECKING_PRECISION 	 = 4, 
	DEPTH 					 = 16
)

	//////////////////INPUTS AND OUTPUTS//////////////

(
	input 																  clk,
	input 																  rst_n,
	input 	  [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] data_in,
	input 																  data_in_available,
	input 																  write_new_data,
	input 																  first,
	input 																  last,
	output reg [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] data_out,
	output reg 															  data_out_available,
	output reg 															  first_out,
	output reg 															  last_out,
	output reg 															  IMF_out_available
);

	//////////////////REGISTERS///////////////////////

	reg																	  read_out;
	reg																	  rd_ptr_updated;
	reg 																	  wr_ptr_updated;
	reg 		  [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] memory [0 : DEPTH - 1];
	reg 											[clogb2(DEPTH)-1:0] wr_ptr;
	reg 											[clogb2(DEPTH)-1:0] rd_ptr;
	reg 											[clogb2(DEPTH)-1:0] ptr_difference;

	//////////////////WIRES///////////////////////////

	wire 																	  already_IMF;
	wire 																	  data_in_available_IMF;
	wire 											[clogb2(DEPTH)-1:0] abs_ptr_difference;
	wire 		  [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] data_minus_avg_spline;
	wire 		  [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] IMF_check_data;

	//////////////////MODULES/////////////////////////

	IMF_check #(
					DATA_SIZE_BITS, 
					PRECISION_BITS_MEMORY, 
					CHECKING_PRECISION
				  )
					
					IMF_c
					
				  (
					clk, 
					IMF_check_data, 
					data_in_available_IMF, 
					first, 
					last,
					already_IMF, 
					!rst_n
				  );


	//////////////////COMBINATORIAL///////////////////

	assign data_minus_avg_spline = $signed(memory[wr_ptr]) - $signed(data_in);
	assign abs_ptr_difference	  = ptr_difference[clogb2(DEPTH)-1] ? -ptr_difference : ptr_difference;
	assign data_in_available_IMF = write_new_data | data_in_available;
	assign IMF_check_data  		  = (write_new_data) ? data_in : data_minus_avg_spline;


	//////////////////SEQUENTIAL///////////////////////////

	//-----------------------------------------------------
	//						READ_OUT
	//-----------------------------------------------------
	always @(posedge clk or negedge rst_n)begin
		if(!rst_n)
			read_out <= 0;
		else if (already_IMF)
			read_out <= 1;
		else if (rd_ptr == DEPTH-1)
			read_out <= 0;
	end

	//-----------------------------------------------------
	//						FIRST_OUT
	//-----------------------------------------------------
	always @(posedge clk or negedge rst_n)begin
		if(!rst_n)
			first_out <= 0;
		else
			first_out <= (((rd_ptr+1 != wr_ptr) & (rd_ptr != wr_ptr) & (rd_ptr == 0)) | (read_out & rd_ptr == 0));
	end

	//-----------------------------------------------------
	//						LAST_OUT
	//-----------------------------------------------------
	always @(posedge clk or negedge rst_n)begin
		if(!rst_n)
			last_out <= 0;
		else
			last_out <= (((rd_ptr+1 != wr_ptr) & (rd_ptr != wr_ptr) & (rd_ptr == DEPTH-1)) | (read_out & rd_ptr == DEPTH-1));
	end

	//-----------------------------------------------------
	//						MEMORY
	//-----------------------------------------------------
	always @(posedge clk)begin
		if(write_new_data)
			memory[wr_ptr] <= data_in;
		else if (data_in_available & !already_IMF & !read_out)
			memory[wr_ptr] <= data_minus_avg_spline;
	end

	//---------------------------------------------------
	//						DATA_OUT
	//---------------------------------------------------
	always @(posedge clk or negedge rst_n) begin
		if(!rst_n)
			data_out <= 0;
		else
			data_out <= memory[rd_ptr];
	end

	//---------------------------------------------------
	//						DATA_OUT_AVAILABLE 
	//---------------------------------------------------
	always @(posedge clk or negedge rst_n) begin
		if(!rst_n)
			data_out_available <= 0;
		else
			data_out_available <= (((rd_ptr+1 != wr_ptr) & (rd_ptr != wr_ptr)) | (read_out & rd_ptr < DEPTH));
	end

	//-----------------------------------------------------
	//						POINTER_DIFFERENCE
	//-----------------------------------------------------
	always@(posedge clk or negedge rst_n) begin
		if(!rst_n)
			ptr_difference = 0; 
		else if(data_in_available | write_new_data) 
			ptr_difference = ptr_difference +1;
		else if((rd_ptr+1 != wr_ptr) & (rd_ptr != wr_ptr))
			ptr_difference = ptr_difference-1;
	end

	//-----------------------------------------------------
	//						READ_POINTER
	//-----------------------------------------------------
	always @(posedge clk or negedge rst_n or posedge already_IMF) begin
		if(!rst_n | already_IMF)
			rd_ptr <= 0;
		else if(((rd_ptr+1 != wr_ptr) & (rd_ptr != wr_ptr)) | (read_out & rd_ptr < DEPTH))
			rd_ptr <= rd_ptr +1;
	end

	//-----------------------------------------------------
	//						READ_POINTER_UPDATED 
	//-----------------------------------------------------
	always @(posedge clk or negedge rst_n) begin
		if(!rst_n)
			rd_ptr_updated <= 0;
		else
			rd_ptr_updated <= (((rd_ptr+1 != wr_ptr) & (rd_ptr != wr_ptr)) | (read_out & rd_ptr < DEPTH));
	end 

	//-----------------------------------------------------
	//						WRITE_POINTER
	//-----------------------------------------------------
	always @(posedge clk or negedge rst_n ) begin
		if(!rst_n)
			wr_ptr <= 0;
		else if(data_in_available | write_new_data)
			wr_ptr <= wr_ptr +1;
	end

	//-----------------------------------------------------
	//						WRITE_POINTER_UPDATED  
	//-----------------------------------------------------
	always @(posedge clk or negedge rst_n) begin
		if(!rst_n)
			wr_ptr_updated <= 0;
		else
			wr_ptr_updated <= data_in_available | write_new_data;
	end 

	//-----------------------------------------------------
	//						IMF_OUT_AVAILABLE
	//-----------------------------------------------------
	always @(posedge clk or negedge rst_n) begin
		if(!rst_n)
			IMF_out_available <= 0;
		else
			IMF_out_available <= read_out;
	end 

	//-----------------------------------------------------
	//	Function for calculating number of bits needed to cover DEPTH.
	//-----------------------------------------------------
		function integer clogb2;
		input [31:0] depth;
			begin
				depth = depth-1;
				for(clogb2 = 0; depth>0; clogb2 = clogb2 +1)
					depth = depth >> 1;
			end
		endfunction

endmodule
