

module calculate_second_derivative 

	//////////////////PARAMETERS///////////////////////////

#(
	parameter 
	DATA_SIZE_BITS 	= 16, 
	PRECISION_BITS_CS = 16,
	DEPTH 				= 512
)

	//////////////////INPUTS AND OUTPUTS////////////////////

(
	input																		clk,
	input 																	rst_n,
	input signed 							[PRECISION_BITS_CS-1:0] C_prime,
	input signed	 [(DATA_SIZE_BITS + PRECISION_BITS_CS):0] D_prime,
	input																		first,
	input																		last,
	input																		data_in_available,
	output reg 		 [(DATA_SIZE_BITS + PRECISION_BITS_CS):0] y_dd,
	output reg 																first_out,
	output reg 																last_out,
	output reg 																data_out_available
);

	//////////////////REGISTERS/////////////////////////////

	reg 			  [(DATA_SIZE_BITS + PRECISION_BITS_CS)-1:0] D_prime_memory [DEPTH/2-1:0]; 
	reg 									 [(PRECISION_BITS_CS)-1:0] C_prime_memory [DEPTH/2-1:0];
	reg 			[(DATA_SIZE_BITS + PRECISION_BITS_CS+1)-1:0] y_dd_memory  	[DEPTH/2-1:0];
	reg 																		back_sub_delayed;

	reg 										  [clogb2(DEPTH/2)-1:0] pointer;
	reg 										  [clogb2(DEPTH/2)-1:0] rd_ptr_y_dd;
	reg 											 [clogb2(DEPTH)-1:0] counter;
	reg 																		read_out;
	reg																		back_sub;
	reg 				 						  [clogb2(DEPTH/2)-1:0] extreme_counter;
	reg 										 	 [clogb2(DEPTH)-1:0] number_of_extremes;

	//////////////////WIRES////////////////////////////////	
	
	wire signed   	 [(DATA_SIZE_BITS + PRECISION_BITS_CS):0] D_m_C_prime_t_y_dd;
	wire signed [(DATA_SIZE_BITS + 2*PRECISION_BITS_CS)-1:0] C_prime_y_dd;
	
	//////////////////COMBINATORIAL////////////////////////
	
	assign C_prime_y_dd = ($signed(C_prime_memory[pointer])*$signed(y_dd_memory[pointer+1])) >>> PRECISION_BITS_CS;

	//////////////////SEQUENTIAL///////////////////////////	
	
	//-----------------------------------------------------
	//						memory
	//-----------------------------------------------------
	always@(posedge clk) begin
		if(data_in_available)
			C_prime_memory[pointer] = C_prime;
	end

	always@(posedge clk) begin
		if(data_in_available)
			D_prime_memory[pointer] = D_prime;
	end

	always@(posedge clk) begin
		if(back_sub)begin
			if(pointer == number_of_extremes-3)
				y_dd_memory[pointer] = $signed(D_prime_memory[pointer]);
			else 
				y_dd_memory[pointer] = $signed(D_prime_memory[pointer]) - $signed(C_prime_y_dd[(DATA_SIZE_BITS + PRECISION_BITS_CS+1)-1:0]);
		end
	end

	always@(posedge clk or negedge rst_n) begin
		if (!rst_n)
			y_dd <= 0;
		else if(read_out)
			y_dd <= y_dd_memory[rd_ptr_y_dd];
	end

	
	//-----------------------------------------------------
	//						pointer
	//-----------------------------------------------------
	always@(posedge clk or negedge rst_n) begin
		if(!rst_n)
			pointer <= 0;
		else if(back_sub & (counter != number_of_extremes-3))
			pointer <= pointer-1;
		else if(data_in_available & !last)
			pointer <= pointer +1;
	end

	
	//-----------------------------------------------------
	//						rd_ptr_y_dd
	//-----------------------------------------------------
	always@(posedge clk or negedge rst_n) begin
		if(!rst_n)
			rd_ptr_y_dd <= 0;
		else if(read_out)
			rd_ptr_y_dd <= rd_ptr_y_dd + 1;
		else
			rd_ptr_y_dd <= 0;

	end

	
	//-----------------------------------------------------
	//						controll variables
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n) begin
			back_sub <= 0;
			read_out <= 0;
		end
		else if(last) begin
			back_sub <= 1;
			read_out <= 0;
		end
		else if(counter == number_of_extremes-3) begin
			back_sub <= 0;
			read_out <= 1;
		end
		else if(counter == 2*number_of_extremes-5) begin
			back_sub <= 0;
			read_out <= 0;
		end
		
	end


	//-----------------------------------------------------
	//						first_out
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			first_out <= 0;
		else
			first_out <= (counter == (number_of_extremes-2)) & read_out;
	end

	
	//-----------------------------------------------------
	//						last_out
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			last_out <= 0;
		else
			last_out <= (counter == 2*number_of_extremes-5);	
	end

	
	//-----------------------------------------------------
	//						counter
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			counter <= 0;
		else if(back_sub | read_out)
			counter <= counter+1;	
		else
			counter <= 0;
	end

	
	//-----------------------------------------------------
	//						data_out_available
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			data_out_available <= 0;
		else
			data_out_available <= read_out;	
	end

	
	//-----------------------------------------------------
	//						extreme_counter
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			extreme_counter <= 0;
		else if(last)
			extreme_counter <= 0;
		else if(data_in_available)
			extreme_counter <= extreme_counter+1;	
	end

	
	//-----------------------------------------------------
	//						number_of_extremes
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			number_of_extremes <= 0;
		else if(last)
			number_of_extremes <= extreme_counter +3;	
	end

	
	//-----------------------------------------------------
	//			Function for calculating floor(log2(N-1))
	//-----------------------------------------------------
		function integer clogb2;
		input [31:0] depth;
			begin
				depth = depth-1;
				for(clogb2 = 0; depth>0; clogb2 = clogb2 +1)
					depth = depth >> 1;
			end
		endfunction

endmodule


